import argparse
import os
import sys
import random
import signal
from random import shuffle

from PyQt5.QtCore import Qt, QObject, pyqtSignal, QTimer
from PyQt5.QtGui import QPen, QKeySequence
from PyQt5.QtWidgets import QApplication, QInputDialog, QShortcut, QProgressDialog, QMessageBox, QPushButton, QMenu
from PyQt5.QtWidgets import QMainWindow, QHBoxLayout, QWidget, QTreeWidgetItem, QSplitter, QGraphicsRectItem, QTabWidget, QLabel, QVBoxLayout, \
    QSizePolicy, QFormLayout, QCheckBox, QGroupBox
from easyconfig2.easyconfig import EasyConfig2
from numpy.ma.core import maximum
from swikv4.pages.swik_page import SwikPage

from qrgrader.dialogs import ControlDialog
from swikv4.widgets.swik_basic_widget import SwikBasicWidget

from qrgrader.filter_dialog import FilterDialog
from qrgrader.utils import makedir
from qrgrader.code import Code
from qrgrader.code_set import CodeSet
from qrgrader.common import check_workspace, get_workspace_paths, Questions, Nia, StudentsData, get_prefix
from qrgrader.pdf_tree import PDFTree, NumericTreeWidgetItem
from qrgrader.rubric import Rubric


class Mark(QGraphicsRectItem):
    class Signal(QObject):
        double_click = pyqtSignal(object)

    def __init__(self, code):
        super().__init__()
        self.signal = Mark.Signal()
        self.code = code
        self.setRect(code.x, code.y, code.w, code.h)
        if code.marked:
            self.setPen(QPen(Qt.red, 2))
        else:
            self.setPen(QPen(Qt.transparent, 2))

    def mouseDoubleClickEvent(self, event):
        self.signal.double_click.emit(self.code)

class EditableLabel(QLabel):
    new_value = pyqtSignal(int)

    def __init__(self):
        super().__init__()

    def mouseDoubleClickEvent(self, event):
        super().mouseDoubleClickEvent(event)
        a, b = QInputDialog.getInt(self, "Edit Value", "Enter new value:", int(self.text().replace("X","0").replace("Y","0")))
        if b:
            self.setText(str(a))
            self.new_value.emit(a)

class MainWindow(QMainWindow):
    def __init__(self, schema_filenames, randomize):
        super().__init__()
        makedir(os.path.expanduser("~") + os.sep + ".config/qrgrader/")

        self.randomize = randomize
        self.config = EasyConfig2(filename=os.path.expanduser("~") + os.sep + ".config/qrgrader/qrgrader.yaml")
        self.cfg_geometry = self.config.root().addPrivate("geometry", default=[0, 0, 1200, 1000, False])
        self.cfg_buttons_height = self.config.root().addPrivate("buttons_height")
        self.cfg_buttons_font = self.config.root().addPrivate("buttons_font")
        self.config.load()

        self.current_exam = None
        self.detected = CodeSet()
        self.type_a = None
        self.type_n = None
        self.changed = CodeSet()

        # Rubrics
        self.rubrics = []
        self.rubrics_files = schema_filenames
        self.rubrics_labels = []
        self.rubrics_cb = []

        (self.dir_workspace,
         self.dir_data,
         self.dir_scanned, _,
         self.dir_xls,
         self.dir_publish, _) = get_workspace_paths(os.getcwd())

        self.prefix = get_prefix()
        self.xls_questions = Questions(self.dir_xls + self.prefix + "questions.csv")
        self.xls_nia = Nia(self.dir_xls + self.prefix + "nia.csv")
        self.xls_data = StudentsData(self.dir_xls + self.prefix + "data.csv")

        self.central_widget = QWidget()
        self.main_layout = QVBoxLayout()
        self.central_widget.setLayout(self.main_layout)
        self.setCentralWidget(self.central_widget)
        self.pdf_tree = PDFTree()
        self.pdf_tree.setColumnCount(4)
        self.swik = SwikBasicWidget()
        self.swik.view.document_ready.connect(self.load_finished)
        self.splitter = QSplitter()

        # Prepare Details layout
        self.name_lbl = QLabel()
        self.name_lbl.setMinimumWidth(300)
        self.name_lbl.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)

        self.nia_lbl = EditableLabel()
        self.nia_lbl.new_value.connect(self.change_nia)
        self.nia_lbl.setMinimumWidth(100)
        self.nia_lbl.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)

        self.group_lbl = QLabel()
        self.group_lbl.setMinimumWidth(100)
        self.group_lbl.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)

        details_layout = QHBoxLayout()
        details_layout.setAlignment(Qt.AlignLeft)
        details_layout.addWidget(QLabel("Name:"))
        details_layout.addWidget(self.name_lbl)
        details_layout.addWidget(QLabel("NIA:"))
        details_layout.addWidget(self.nia_lbl)
        details_layout.addWidget(QLabel("Group:"))
        details_layout.addWidget(self.group_lbl)

        self.main_layout.addWidget(self.splitter)
        self.main_layout.addLayout(details_layout)

        self.rubrics_tabs = QTabWidget()

        helper = QWidget()
        helper.setLayout(QVBoxLayout())

        self.scores_layout = QFormLayout()
        framebox = QGroupBox("Scores")
        framebox.setLayout(self.scores_layout)

        helper.layout().addWidget(framebox)
        helper.layout().addWidget(self.rubrics_tabs)

        self.quiz_score_lbl = QLabel("0")
        self.total_score_lbl = QLabel("0")
        self.quiz_cb = QCheckBox("Quiz")
        self.quiz_cb.setChecked(True)
        self.quiz_cb.stateChanged.connect(self.score_checkbox_changed)
        self.scores_layout.addRow(self.quiz_cb, self.quiz_score_lbl)
        self.scores_layout.addRow("Total: ", self.total_score_lbl)

        done_info_helper = QWidget()
        done_info_helper.setLayout(QVBoxLayout())
        self.number_assesed_lbl = QLabel("")
        done_info_helper.layout().addWidget(self.number_assesed_lbl)
        done_info_helper.layout().addWidget(self.pdf_tree)
        done_info_helper.setMaximumWidth(280)

        self.splitter.addWidget(done_info_helper)
        self.splitter.addWidget(self.swik)
        self.splitter.addWidget(helper)

        self.setWindowTitle("QR Grader")
        self.shortcut = QShortcut(QKeySequence('Ctrl+F'), self)
        self.shortcut.activated.connect(self.find_people)
        self.show()


        files = os.listdir(self.dir_publish)
        files = [f for f in files if f.endswith(".pdf") and f.replace(".pdf", "").isdigit()]

        progress = QProgressDialog("Processing...", None, 0, 10 + len(files), self)
        progress.setWindowModality(Qt.WindowModal)
        progress.setWindowTitle("QRGrader")
        progress.show()

        def delayed():
            # Show the progress bar
            self.load_tables()
            self.load_detected()
            self.load_schemas()
            progress.setValue(10)

            self.populate_pdf_tree(self.randomize)
            self.pdf_tree.currentItemChanged.connect(self.pdf_tree_selection_changed)

            if self.pdf_tree.topLevelItemCount() > 0:
                self.pdf_tree.setCurrentItem(self.pdf_tree.topLevelItem(0))

            for index in range(self.pdf_tree.topLevelItemCount()):
                progress.setValue(10 + index)
                item = self.pdf_tree.topLevelItem(index)
                exam_id = int(item.text(1))

                # Update scores
                score = self.get_full_score(exam_id)
                item.setText(3, str(round(score,2)))

                # Check if multiple marks or bad NIA
                nia = self.xls_nia.get_nia(exam_id)
                if len(self.get_multiple_marks(exam_id)) > 0:
                    item.setText(2, "!")
                elif nia is None or type(nia) == str:
                    item.setText(2, "@")
                # elif self.get_missing_pq_marks(exam_id):
                #     item.setText(2, "P")
                else:
                    item.setText(2, "")

                if len(self.rubrics) > 0:
                    self.update_done_color(item)
                    self.update_number_assessed()

            self.rubrics_tabs.currentChanged.connect(self.rubric_tab_changed)

            progress.hide()

            # add shortcut to find people
        QTimer.singleShot(100, delayed)

    def contextMenuEvent(self, event):
        pos = self.swik.view.mapFrom(self, event.pos())
        item = self.swik.view.get_item_at_position(pos,SwikPage)
        if item is not None:
            menu = QMenu(self)
            move_action = menu.addAction("Move Codes")
            action = menu.exec_(self.mapToGlobal(event.pos()))
            if action == move_action:
                dialog = ControlDialog(item.index, lambda: self.detected.save(self.dir_data + self.prefix + "detected.csv"),
                                        lambda: self.move_codes(item.index,0, -2),
                                        lambda: self.move_codes(item.index,0, 2),
                                        lambda: self.move_codes(item.index,  -2, 0),
                                        lambda: self.move_codes(item.index,2, 0),
                                        lambda: self.move_codes(item.index,0, 0, 1.01),
                                        lambda: self.move_codes(item.index,0, 0, 0.99), self)
                dialog.show()


    def rubric_tab_changed(self, index):
        rubric = self.rubrics_tabs.currentWidget()
        if rubric is not None:
            index = rubric.get_page()
            try:
                self.swik.view.move_to_page(index)
            except Exception as e:
                pass

    def update_done_color(self, item):
        howmany = 0
        for rubric in self.rubrics:
            if rubric.assessed(int(item.text(1))):
                howmany += 1
        if howmany == 0:
            item.setForeground(1, Qt.red)
            symb = ""
        elif howmany < len(self.rubrics):
            item.setForeground(1, Qt.magenta)
            symb = "⏳"
        else:
            item.setForeground(1, Qt.black)
            symb = "✓"

        if item.text(2) not in ["!", "@"]:
            item.setText(2, symb)

    def update_number_assessed(self):
        rubric = self.rubrics_tabs.currentWidget()

        if rubric is not None:
            count = 0
            for index in range(self.pdf_tree.topLevelItemCount()):
                item = self.pdf_tree.topLevelItem(index)
                if rubric.assessed(int(item.text(1))):
                    count += 1
            self.number_assesed_lbl.setText(f"Done: {count}/{self.pdf_tree.topLevelItemCount()} ({count * 100 / self.pdf_tree.topLevelItemCount():.2f}%)")

    def show(self):
        x, y, w, h, fullscreen = self.cfg_geometry.get()
        if fullscreen:
            self.showFullScreen()
        else:
            self.setGeometry(x, y, w, h)
        super().show()


    def change_nia(self, nia):
        self.xls_nia.set_nia(self.current_exam, nia)
        self.xls_nia.save()
        item = self.pdf_tree.currentItem()
        exam_id = int(item.text(1))
        nia = self.xls_nia.get_nia(exam_id)
        if len(self.get_multiple_marks(exam_id)) > 0:
            item.setText(2, "!")
        elif nia is None or type(nia) == str:
            item.setText(2, "@")
        else:
            item.setText(2, "")

    def find_people(self):
        text, ok = QInputDialog.getText(self, "Find People", "Enter NIA or Name:")
        if ok:
            nia = self.xls_data.get_nia_from_name(text)
            if nia is None:
                nia = int(text) if text.isdigit() else -1
            exam_id = self.xls_nia.get_exam(nia)
            for index in range(self.pdf_tree.topLevelItemCount()):
                item = self.pdf_tree.topLevelItem(index)
                if int(item.text(1)) == exam_id:
                    self.pdf_tree.setCurrentItem(item)
                    self.pdf_tree.scrollToItem(item)
                    break
            else:
                QMessageBox.information(self, "Not Found", "No exam found for NIA or Name: " + text)

    def closeEvent(self, a0):
        self.cfg_geometry.set(
            [self.geometry().x(), self.geometry().y(), self.geometry().width(), self.geometry().height(),
             self.isFullScreen()])
        self.config.save()

    def load_detected(self):
        self.detected.load(self.dir_data + self.prefix + "detected.csv")
        # Pre-select codes to improve performance
        # type_a and type_n contain the codes pointer so
        # any modification in them will be reflected in detected
        self.type_a = self.detected.select(type=Code.TYPE_A)
        self.type_n = self.detected.select(type=Code.TYPE_N)

        self.type_p = self.detected.select(type=Code.TYPE_P)
        self.type_q = self.detected.select(type=Code.TYPE_Q)


    def load_tables(self):

        if not self.xls_questions.load():
            print("ERROR: questions.csv file not present")
            sys.exit(1)

        if not self.xls_nia.load():
            print("ERROR: nia.csv file not present")
            sys.exit(1)

        if not self.xls_data.load():
            print("WARNING: data.csv file not present")

    def rubric_filtered(self):
        for index in range(self.pdf_tree.topLevelItemCount()):
            item = self.pdf_tree.topLevelItem(index)
            exam_id = int(item.text(1))
            comply = True
            for rubric in self.rubrics:
                comply = comply and rubric.comply_with_filter(exam_id)
                item.setHidden(not comply)

    def load_schemas(self):
        for filename in self.rubrics_files:
            name = os.path.basename(filename).replace(".scm", "")
            r1 = Rubric(filename, self.dir_xls, buttons_height=self.cfg_buttons_height.get(), buttons_font=self.cfg_buttons_font.get())
            r1.score_changed.connect(self.rubric_score_changed)
            r1.button_or_value_changed.connect(self.rubric_button_or_value_changed)
            r1.goto_next.connect(self.goto_next)
            r1.filtered.connect(self.rubric_filtered)

            self.rubrics_tabs.addTab(r1, name)
            self.rubrics.append(r1)

            label = QLabel("0")
            rubric_cb = QCheckBox(name + ":")
            rubric_cb.setChecked(True)
            rubric_cb.stateChanged.connect(self.score_checkbox_changed)

            self.rubrics_labels.append(label)
            self.rubrics_cb.append(rubric_cb)

            self.scores_layout.insertRow(1, rubric_cb, label)

    def goto_next(self):
        current = self.pdf_tree.currentItem()
        if current is not None:
            index = self.pdf_tree.indexOfTopLevelItem(current)
            if index < self.pdf_tree.topLevelItemCount() - 1:
                self.pdf_tree.setCurrentItem(self.pdf_tree.topLevelItem(index + 1))

    def score_checkbox_changed(self, state):
        self.update_all_pdf_tree_scores()
        self.update_scores_layout()

    def rubric_score_changed(self, rubric, exam_id):
        self.update_scores_layout()
        self.update_pdf_tree_score()

    def rubric_button_or_value_changed(self):
        # dialog with progress bar no advancement
        progress = QProgressDialog("Processing...","",0, 0)
        progress.setWindowModality(Qt.WindowModal)
        progress.setCancelButton(None)
        progress.setWindowTitle("QRGrader")
        progress.show()
        def delayed():
            self.update_all_pdf_tree_scores()
            self.update_scores_layout()
            progress.hide()
        QTimer.singleShot(100, delayed)

    def update_scores_layout(self):

        total = 0
        quiz_score, full_score = self.get_quiz_score(self.current_exam)
        if full_score != 0:
            self.quiz_score_lbl.setText("<b>" + str(quiz_score) + "</b>/" + str(round(full_score,1)) + " (" + str(round(10*quiz_score / full_score, 2)) + "/10)")
        else:
            self.quiz_score_lbl.setText("<b>0</b>")

        if self.quiz_cb.isChecked():
            total += quiz_score

        for index, r in enumerate(self.rubrics):
            value, full_value = r.compute_score(self.current_exam)
            self.rubrics_labels[index].setText("<b>" + str(value) + "</b>/" + str(full_value) + " (" + str(round(10*value / full_value, 2)) + "/10)")
            if self.rubrics_cb[index].isChecked():
                total += value

        self.total_score_lbl.setText("<b>" + str(round(total,2)) + "</b>")
        return total

    def pdf_tree_selection_changed(self, current, previous):
        self.pdf_tree.set_enabled(False)
        if previous is not None:
            if len(self.rubrics) > 0:
                self.update_done_color(previous)
                self.update_number_assessed()

            for rubric in self.rubrics:
                rubric.push(int(previous.text(1)))

        self.current_exam = int(current.text(1))

        ratio = self.swik.view.get_ratio()
        rubric = self.rubrics_tabs.currentWidget()
        index = rubric.get_page() if rubric is not None else 0



        self.swik.open(f"{self.dir_publish}{self.current_exam}.pdf", ratio=ratio, index=index)

    def load_finished(self):

        self.process_exam()

        nia = self.xls_nia.get_nia(self.current_exam)
        self.nia_lbl.setText(str(nia))
        self.name_lbl.setText(str(self.xls_data.get_name(nia)))
        self.group_lbl.setText(str(self.xls_data.get_group(nia)))

        for rubric in self.rubrics:
            rubric.pull(self.current_exam)

        self.update_scores_layout()
        self.pdf_tree.set_enabled(True)


    def populate_pdf_tree(self, randomize=True):
        files = os.listdir(self.dir_publish)
        #shuffle files

        files = sorted([f.replace(".pdf", "") for f in files if f.endswith(".pdf") and f.replace(".pdf", "").isdigit()])
        shuffle(files)
        for f in files:
            item = NumericTreeWidgetItem(["", f, ""])
            self.pdf_tree.addTopLevelItem(item)

        if not randomize:
            self.pdf_tree.sortByColumn(1, Qt.AscendingOrder)
        self.pdf_tree.renumber()

        # Not updating the scores here, as it is done somewhere else
        # self.update_all_pdf_tree_scores()

    def process_exam(self):
        marks = [x for x in self.swik.view.items() if isinstance(x, Mark)]

        while len(marks) > 0:
            mark = marks.pop()
            self.swik.view.scene().removeItem(mark)

        marks = {}
        exam_id = self.current_exam % 1000

        for index in range(self.swik.renderer.get_document_length()):
            page_codes_type_a = self.type_a.select(exam=exam_id, pdf_page=index + 1)
            #type_a = page_codes.select(type=Code.TYPE_A)
            for code in page_codes_type_a:
                r = Mark(code)
                r.signal.double_click.connect(self.code_clicked)
                r.setParentItem(self.swik.view.get_page(index))
                marks[code] = r
                if code.marked:
                    if self.xls_questions.get_value(code.question, code.answer) > 0:
                        r.setPen(QPen(Qt.green, 2))
                    else:
                        r.setPen(QPen(Qt.red, 2))

            page_codes_type_n = self.type_n.select(exam=exam_id, pdf_page=index + 1)
            for code in page_codes_type_n:
                r = Mark(code)
                r.signal.double_click.connect(self.code_clicked)
                r.setParentItem(self.swik.view.get_page(index))
                if code.marked:
                    r.setPen(QPen(Qt.magenta, 2))

        yellow = self.get_multiple_marks(exam_id)
        for answer in yellow:
            marks[answer].setPen(QPen(Qt.yellow, 2))

    def get_full_score(self, exam_id):
        total = 0
        for index, r in enumerate(self.rubrics):
            if self.rubrics_cb[index].isChecked():
                rubric_value, _ = r.compute_score(exam_id)
                total += rubric_value
        if self.quiz_cb.isChecked():
            quiz_score,quiz_full_value = self.get_quiz_score(exam_id)
            total += quiz_score

        return total

    def update_all_pdf_tree_scores(self):
        for index in range(self.pdf_tree.topLevelItemCount()):
            item = self.pdf_tree.topLevelItem(index)
            score = self.get_full_score(int(item.text(1)))
            item.setText(3, str(round(score,2)))

    def update_pdf_tree_score(self):
        for index in range(self.pdf_tree.topLevelItemCount()):
            item = self.pdf_tree.topLevelItem(index)
            if int(item.text(1)) == self.current_exam:
                score = self.get_full_score(self.current_exam)
                item.setText(3, str(round(score,2)))
                break

    def get_quiz_score(self, exam_id):
        score = 0
        full_value = 0
        exam_codes = self.type_a.select(exam=exam_id % 1000)
        for code in exam_codes:
            value = self.xls_questions.get_value(code.question, code.answer)
            full_value += value if value > 0 else 0
            if code.marked:
                score += value
        score = max(score, 0)  # Ensure score is not negative
        score = round(score, 4) # Round to 4 decimal places
        return score, full_value

    def get_multiple_marks(self, exam_id):
        yellow = []
        exam_id = exam_id % 1000
        type_a = self.type_a.select(exam=exam_id)
        for question in type_a.get_questions():
            answers = type_a.select(question=question)
            marked = sum([1 for x in answers if x.marked])
            if marked > 1:
                for answer in answers:
                    if answer.marked:
                        yellow.append(answer)
        return yellow

    def get_missing_pq_marks(self, exam_id):
        exam_id = exam_id % 1000
        type_p = self.type_p.select(exam=exam_id, marked=1)
        type_q = self.type_q.select(exam=exam_id, marked=1)
        return len(type_p) != len(type_q)


    def code_clicked(self, code):
        self.changed.append(code)
        self.changed.save(self.dir_data + self.prefix + "changed.csv")
        code.marked = not code.marked
        self.process_exam()
        self.update_scores_layout()
        self.update_pdf_tree_score()
        self.detected.save(self.dir_data + self.prefix + "detected.csv")

        item = self.pdf_tree.currentItem()
        nia = self.xls_nia.get_nia(code.unique)

        if len(self.get_multiple_marks(int(code.exam))) > 0:
            item.setText(2, "!")
        elif nia is None or type(nia) == str:
            item.setText(2, "@")
        else:
            item.setText(2, "")

    def move_codes(self, page, x, y, scale=1):
        page_codes_type_a = self.type_a.select(exam=self.current_exam%1000, pdf_page=page + 1)
        page_codes_type_n = self.type_n.select(exam=self.current_exam%1000, pdf_page=page + 1)
        a_plus_n = page_codes_type_n + page_codes_type_a

        if len(a_plus_n) == 0:
            return

        most_left, selected = 1e6, None

        for code in [c for c in a_plus_n if c.marked]:
            if code.x < most_left:
                selected = code
                most_left = code.x

        if scale!=1 and selected is not None:
            # the selected code maintain its position and the others are scaled around it
            x = selected.x * (1 - scale)
            y = selected.y * (1 - scale)

        for code in a_plus_n:
            code.x = code.x * scale + x
            code.y = code.y * scale + y

        self.process_exam()



def main():
    app = QApplication(sys.argv)

    # Let Python's signal handler run every 500ms
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    if not check_workspace():
        print("ERROR: qrgrader must be run from the workspace root")
        sys.exit(1)

    parser = argparse.ArgumentParser(description='qrgui.py')
    parser.add_argument('-s', '--schema', help='Schema to be used', default=[], action="append")
    parser.add_argument('-c', '--create', help="Create schema if doesn't exist", action="store_true")
    parser.add_argument('-r', '--random', help="Randomize exam order", action="store_true")
    args = vars(parser.parse_args())

    filenames = []
    for schema in args["schema"]:
        if schema.endswith(".yaml"):
            print("WARNING: schema MUST NOT be a yaml file.")
            sys.exit(1)

        filename = schema.replace(".scm", "") + ".scm"

        if not os.path.exists(filename):
            if args["create"]:
                print("Creating schema", filename)
                with open(filename, "w", encoding='utf-8') as f:
                    f.write("{}\n")
            else:
                print(f"ERROR: schema {filename} not found")
                sys.exit(1)
        filenames.append(filename)

    main = MainWindow(filenames, args["random"])
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import csv
import os

import gspread.utils
import yaml
from PyQt5.QtCore import Qt, pyqtSignal, QTimer, QSize
from PyQt5.QtGui import QDrag, QPixmap, QPainter
from PyQt5.QtWidgets import (QListWidget,
                             QAbstractItemView, QListWidgetItem, QMenu, QMessageBox,
                             QInputDialog, QColorDialog, QPushButton, QWidget, QVBoxLayout, QFrame)
from qrgrader.common import get_date

from qrgrader.dialogs import ButtonEditDialog, RubricEditDialog
from qrgrader.buttons import StepButton, Shortcut, Button, TextButton, StateButton, Separator, CutterButton, MultiplierButton
from qrgrader.filter_dialog import FilterDialog


class Rubric(QListWidget):
    score_changed = pyqtSignal(object, int)
    goto_next = pyqtSignal()
    button_or_value_changed = pyqtSignal()
    filtered = pyqtSignal()

    def __init__(self, schema_filename, dir_xls, **kwargs):
        super().__init__()

        self.filters = None
        self.exam_id = None
        self.config = {}
        self.scores = {}
        self.schema_dictionary = {}
        self.schema_filename = schema_filename
        name = self.schema_filename.split(".")[0]
        self.scores_filename = name + ".yaml"

        self.xls_filename = dir_xls + get_date() + "_" + name + ".csv"
        self.current_exam_id = None
        self.modified = False
        self.buttons_height = kwargs.get("buttons_height")
        self.buttons_font = kwargs.get("buttons_font")

        self.setDragDropMode(QAbstractItemView.InternalMove)
        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDropIndicatorShown(True)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.setMinimumWidth(115)

        self.schemaChanged.connect(self.save_schema) # type: ignore
        self.customContextMenuRequested.connect(self.button_list_right_click)

        self.filter_btn = QPushButton("Apply Filter")
        self.filter_btn.setFlat(True)
        self.filter_btn.setCheckable(True)
        self.filter_btn.setStyleSheet("background-color: #F0A0A0")
        self.filter_btn.clicked.connect(self.toggle_filter)

        container = QWidget()
        layout = QVBoxLayout(container)
        layout.setContentsMargins(5, 5, 5, 0)  # margins you want
        layout.addWidget(self.filter_btn)
        item = QListWidgetItem()
        self.addItem(item)
        item.setSizeHint(container.sizeHint())
        self.setItemWidget(item, container)

        line = QFrame()
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        item = QListWidgetItem()
        item.setFlags(Qt.NoItemFlags)  # optional: makes it non-selectable
        item.setSizeHint(line.sizeHint())
        self.addItem(item)
        self.setItemWidget(item, line)


        self.populate()
        self.load_scores()

    def toggle_filter(self):
        if not self.filter_btn.isChecked():
            self.filters = None
            self.filtered.emit()
        else:
            filter_dialog = FilterDialog(self)
            if filter_dialog.exec() == QInputDialog.Accepted:
                self.set_filter(filter_dialog.get_selection())
                self.filtered.emit()
            else:
                self.filter_btn.setChecked(False)
                self.filters = None
                self.filtered.emit()


    def set_filter(self, filters):
        self.filters = filters
        self.filtered.emit()

    def comply_with_filter(self, exam_id):
        if self.filters is None:
            return True

        exam_scores = self.scores.get(exam_id, {})

        ok = True
        for k, v in self.filters.items():
            button_score = exam_scores.get(k, {})
            print("Filtering:", k, v, button_score)
            ok = ok and button_score.get("value", -1) == v.get("value")
        return ok

    def load_scores(self):
        if os.path.exists(self.scores_filename):
            with open(self.scores_filename) as file:
                self.scores = yaml.safe_load(file)

    def push(self, exam_id):
        if exam_id is not None:
            if self.store(exam_id):
                self.save_scores()

    def pull(self, exam_id):
        self.exam_id = exam_id
        self.retrieve(self.exam_id)

    # def exam_changed(self, exam_id, prev_exam_id):
    #     self.exam_id = exam_id
    #     if prev_exam_id is not None:
    #         self.store(prev_exam_id)
    #     self.save_scores()
    #
    #     self.retrieve(self.exam_id)

    def get_page(self):
        return self.config.get("page", 1) - 1

    def keyPressEvent(self, e):
        if e.key() == Qt.Key_Down:
            self.goto_next.emit()
        else:
            super().keyPressEvent(e)

    def populate(self):
        if os.path.exists(self.schema_filename):
            # Load schema
            with open(self.schema_filename, newline='') as csvfile:
                content = yaml.full_load(csvfile)

                self.config = content.get("config", {})
                self.schema_dictionary = content.get("buttons", {})

                # Create lateral panel and buttons
                for button_name in self.schema_dictionary:
                    button_config = self.schema_dictionary[button_name]
                    button_config["height"] = self.buttons_height
                    button_config["font"] = self.buttons_font

                    if button_config.get("type") == 'button':
                        button = StepButton(button_name, **button_config)
                        button.score_changed.connect(self.button_clicked)  # type: ignore
                    elif button_config.get("type") == 'text':
                        button = TextButton(button_name, **button_config)
                    elif button_config.get("type") == 'cutter':
                        button = CutterButton(button_name, **button_config)
                        button.score_changed.connect(self.button_clicked)
                    elif button_config.get("type") == 'multiplier':
                        button = MultiplierButton(button_name, **button_config)
                        button.score_changed.connect(self.button_clicked)
                    elif button_config.get("type") == 'shortcut':
                        button = Shortcut(button_name, **button_config)
                        button.clicked.connect(self.shortcut_activated)
                    elif button_config.get("type") == 'separator':
                        button = Separator(button_name, **button_config)
                    else:
                        button = None

                    if button is not None:
                        # Create Item in ListWidget
                        item = QListWidgetItem()
                        self.addItem(item)
                        self.setItemWidget(item, button)
                        item.setSizeHint(button.sizeHint())

    def save_schema(self):
        buttons = {}
        for b in self.filter_buttons(Button):
            buttons[b.get_name()] = b.get_config()

        self.schema_dictionary = buttons

        with open(self.schema_filename, "w", encoding='utf-8') as f:
            schema = {"buttons": buttons, "config": self.config}
            yaml.dump(schema, f, sort_keys=False)

    def button_clicked(self):
        self.store(self.exam_id)
        self.score_changed.emit(self, self.exam_id)
        button = self.sender()
        if button.get_click_next() and button.is_checked():
            QTimer.singleShot(500, self.goto_next.emit)

    def add_button(self):

        button = self.get_dialog()

        if button is None:
            return

        # Create Item in ListWidget
        item = QListWidgetItem()
        self.addItem(item)

        self.setItemWidget(item, button)
        item.setSizeHint(button.sizeHint())
        self.save_schema()
        self.button_or_value_changed.emit()

    def get_dialog(self, button=None):

        dialog = ButtonEditDialog(self, button)

        if not dialog.exec():
            return None

        name, kind, config = dialog.get()

        self.schema_dictionary[name] = config

        if kind == 'button':
            button = StepButton(name, **config)
            button.score_changed.connect(self.button_clicked)
        elif kind == 'text':
            button = TextButton(name, **config)
        elif kind == 'cutter':
            button = CutterButton(name, **config)
            button.score_changed.connect(self.button_clicked)
        elif kind == 'multiplier':
            button = MultiplierButton(name, **config)
            button.score_changed.connect(self.button_clicked)
        elif kind == 'separator':
            button = Separator(name, **config)
        else:
            button = None

        return button

    def edit_button(self, position):
        item = self.item(position)
        widget = self.itemWidget(item)
        prev_name = widget.get_name()

        button = self.get_dialog(widget)

        if button is None:
            return

        self.setItemWidget(item, button)

        # Rename the key in the dictionary
        if prev_name is not None:
            for exam_id in self.scores:
                if prev_name in self.scores[exam_id].keys():
                    self.scores[exam_id][button.get_name()] = self.scores[exam_id].pop(prev_name)

        self.save_schema()
        self.button_or_value_changed.emit()

    def button_list_right_click(self, pos):
        menu = QMenu()
        item = self.item(self.currentRow())
        widget = self.itemWidget(item)

        if isinstance(widget, StateButton):
            menu.addAction("Edit", lambda: self.edit_button(self.currentRow()))
            menu.addAction("Duplicate", lambda: self.duplicate_button(self.currentRow()))
            menu.addAction("Remove", lambda: self.delete_button(self.currentRow()))
            menu.addAction("Add Comment", lambda: self.add_comment(self.currentRow()))
            menu.addSeparator()
        elif isinstance(widget, Shortcut):
            menu.addAction("Remove", lambda: self.remove_shortcut(self.currentRow()))
            menu.addAction("Change Color", lambda: self.set_shortcut_color(self.currentRow()))
            menu.addSeparator()
        menu.addAction("Edit Rubric Config", self.edit_rubric_config)
        menu.addAction("Add Button", self.add_button)
        menu.addAction("Add Shortcut", self.add_shortcut)

        menu.exec(self.mapToGlobal(pos))
        self.clearSelection()
        self.clearFocus()

    def compute_score(self, exam_id):
        this_exam_data = self.scores.get(int(exam_id), {})

        total = 0
        points = 0
        text = "="
        for button in self.filter_buttons(StepButton):
            this_button_data = this_exam_data.get(button.get_name(), {})
            value = this_button_data.get("value", 0)
            if value >= 0:
                value = value * button.get_full_value() / 100.0
                text = text + 'N("' + button.get_name() + '")+' + str(value) + ' + '
                points = points + value

            total = total + button.get_full_value() * button.get_weight()
        text += "0"
        #print(text)
        cut = 1
        for button in self.filter_buttons(CutterButton):
            cut = min(cut, button.get_percent())

        points = min(points, total * cut)

        multiplier = 1
        for button in self.filter_buttons(MultiplierButton):
            multiplier = button.get_percent()

        points = points * multiplier
        if total > 0:
            points = points / total * self.config.get("weight", 10)
        points = round(points, self.config.get("precision", 2))
        return points, self.config.get("weight", 10)

    def save_scores(self):
        with open(self.scores_filename, "w", encoding='utf-8') as file:
            yaml.dump(self.scores, file)

        self.save_xls()

    def save_xls(self):

        with open(self.xls_filename, "w", encoding='utf-8') as f:
            filtered = {
                name: config
                for name, config in self.schema_dictionary.items()
                if config.get("type") not in ['separator', 'shortcut']
            }

            # HEADER
            row = "EXAM_ID\tSCORE"
            for button_name, button_config in filtered.items():
                row += "\t" + button_name
            f.write(row + "\n")

            row = "\tWeight"
            for button_name, button_config in filtered.items():
                row += "\t" + f"{button_config.get("weight", 1)}"

            f.write(row + "\n")

            # SCORES
            row = "\tValue"
            for button_name, button_config in filtered.items():
                if button_config.get("type") == 'button':
                    row += "\t" + "{:.2f}".format(button_config.get("full_value", 1))
                elif button_config.get("type") == 'text':
                    row += "\t "
                else:
                    row += "\t" + "{:.2f}".format(button_config.get("percent", 1))

            f.write(row + "\n")


            col = 3
            cutter = []
            multipliers = []
            normal = []

            for button_name, button_config in filtered.items():
                if button_config.get("type") == 'multiplier':
                    multipliers.append(col)
                elif button_config.get("type") == 'button':
                    normal.append(col)
                elif button_config.get("type") == 'cutter':
                    cutter.append(col)
                col += 1

            # Sort scores by exam_id
            self.scores = dict(sorted(self.scores.items()))

            # VALUES
            current_row = 3
            for exam_id, exam_score_items in self.scores.items():
                current_row += 1

                row = str(exam_id)
                suma = "SUMIF({0"
                row2 = "(0"
                for col in normal:
                    full_value = gspread.utils.rowcol_to_a1(3, col)
                    weight = gspread.utils.rowcol_to_a1(2, col)
                    value = gspread.utils.rowcol_to_a1(current_row, col)
                    row2 += "+" + full_value + "*" + value
                    suma += "," + full_value + "*" + weight
                suma = suma + '},">0")'
                row2 += ")/" + suma


                for col in multipliers:
                    percent = gspread.utils.rowcol_to_a1(2, col)
                    value = gspread.utils.rowcol_to_a1(current_row, col)
                    row2 += "*IF("+value+"=0,1,"+percent+")"


                for col in cutter:
                    percent = gspread.utils.rowcol_to_a1(2, col)
                    value = gspread.utils.rowcol_to_a1(current_row, col)
                    row2 = "if(" + value + " = 0,"+ row2 +", min("+row2+","+percent+"))" #" + percent + "*" + suma + "))"

                row += "\t=" + row2
                for button_name, button_config in filtered.items():

                    button_type = button_config.get("type")
                    button_state = exam_score_items.get(button_name)  # value and comment or text

                    if button_state is not None:
                        if button_type == 'button':
                            value = button_state.get("value")
                            value = " " if value == -1 else round(value / 100, 2)
                        elif button_type == 'text':
                            value = button_state.get("text")
                        else:
                            value = button_state.get("value")
                    else:
                        value = " "

                    row += "\t {:4s}".format(str(value))

                f.write(row + "\n")

    def shortcut_activated(self):
        buttons = self.sender().get_buttons()

        for b in self.filter_buttons(StepButton):  # type: StepButton
            b.button.setChecked(b.get_name() in buttons)
            b.clicked()

    def add_shortcut(self):
        text, ok = QInputDialog.getText(self, "Shortcut", "Name:")
        if ok:

            buttons = []
            for b in self.filter_buttons(StepButton):  # type: StepButton
                if b.button.isChecked():
                    buttons.append(b.get_name())

            b2 = Shortcut(text, buttons=buttons)
            b2.clicked.connect(self.shortcut_activated)  # type: ignore

            item = QListWidgetItem()
            self.addItem(item)
            self.setItemWidget(item, b2)
            item.setSizeHint(b2.sizeHint())
            item.setFlags(item.flags() & ~Qt.ItemIsDragEnabled)
            self.save_schema()

    #
    def delete_button(self, position):
        ret = QMessageBox().question(self, '', "Are you sure?", QMessageBox.Yes | QMessageBox.No)

        if ret == QMessageBox.Yes:
            item = self.takeItem(position)
            del item
            self.save_schema()
            self.button_or_value_changed.emit()

    def add_comment(self, position):
        # help me get text comment with dialog
        item = self.item(position)
        widget = self.itemWidget(item)
        text, ok = QInputDialog.getText(self, 'Add Comment', 'Comment:', text=widget.get_comment())
        if ok:
            item = self.item(position)
            widget = self.itemWidget(item)
            widget.setToolTip(text)
            widget.set_comment(text)

    def duplicate_button(self, position):

        # button = self.get_dialog()
        #
        # if button is None:
        #     return
        #
        # # Create Item in ListWidget
        # item = QListWidgetItem()
        # self.addItem(item)
        #
        # self.setItemWidget(item, button)
        # item.setSizeHint(button.sizeHint())
        # self.save_schema()

        item = self.item(position)
        widget = self.itemWidget(item)
        name, valid = QInputDialog.getText(self, 'Duplicate', 'Name:', text=widget.get_name() + "_copy")
        if valid:
            button = widget.__class__(name, **widget.get_config().copy())
            button.score_changed.connect(self.button_clicked)
            item = QListWidgetItem()
            self.addItem(item)
            self.setItemWidget(item, button)
            item.setSizeHint(button.sizeHint())
            self.save_schema()

    def edit_rubric_config(self):
        dialog = RubricEditDialog(self.config)
        if dialog.exec():
            self.save_schema()
            self.button_or_value_changed.emit()

    def set_shortcut_color(self, position):
        item = self.item(position)
        widget = self.itemWidget(item)
        color = QColorDialog.getColor()
        if color.isValid():
            widget.set_color(color.name())
            self.save_schema()

    def remove_shortcut(self, position):
        self.takeItem(position)
        self.save_schema()

    # def is_done(self, exam_id):
    #     done = False
    #     grades_for_this_exam = self.rubric_grades_data.get(exam_id)
    #     if grades_for_this_exam is None:
    #         return False
    #
    #     for b in self.buttons(StepButton):  # type: Score
    #         this_button = grades_for_this_exam.get(b.get_name(), {})
    #         value = this_button.get("value", -1)
    #         done = done or value != -1
    #     for b in self.buttons(Text):
    #         this_button = grades_for_this_exam.get(b.get_name(), {})
    #         value = this_button.get("text", "")
    #         done = done or value != ""
    #
    #     return done
    #
    def store(self, exam_id):
        assessed = False

        # Remove previous scores for this exam_if in case
        # the exam_id has been "uncorrected"
        self.scores.pop(exam_id, None)
        for b in self.filter_buttons(StateButton):  # type: Score
            state = b.get_state()

            # Only store if a button has been pressed or there is some text
            if (state.get("value", -1) != -1 or
                    state.get("text", "") != "" or
                    state.get("percent", -1) != -1):

                if self.scores.get(exam_id) is None:
                    self.scores[exam_id] = {}
                self.scores[exam_id][b.get_name()] = state
                assessed = True
        return assessed

    def assessed(self, exam_id):
        return self.scores.get(exam_id) is not None



    def retrieve(self, exam_id):
        self.current_exam_id = exam_id
        rubric_data = self.scores.get(exam_id, {})

        if rubric_data is not None:
            for button in self.filter_buttons(StateButton):  # type: Score
                button.blockSignals(True)
                button.set_state(rubric_data.get(button.get_name(), {}))
                button.blockSignals(False)

    def startDrag(self, ev):
        selected = self.selectedIndexes()[0].row()
        item = self.item(selected)
        widget = self.itemWidget(item)
        qd = QDrag(self)
        qd.setMimeData(self.model().mimeData(self.selectedIndexes()))
        pm = QPixmap(400, 20)
        pm.fill(Qt.transparent)
        qp = QPainter(pm)
        qp.drawText(10, 15, widget.get_name())
        qd.setPixmap(pm)
        qd.exec(ev, Qt.MoveAction)
        del qp

    schemaChanged = pyqtSignal()

    def dropEvent(self, event):
        super().dropEvent(event)
        self.schemaChanged.emit()  # type: ignore

    def filter_buttons(self, kind=StepButton):
        buttons = []
        for i in range(self.count()):
            item = self.item(i)
            widget = self.itemWidget(item)
            if isinstance(widget, kind):
                buttons.append(widget)
        return buttons





from cryptography.fernet import Fernet
from cryptography.fernet import Fernet
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.backends import default_backend
import base64
import os
import argparse

def encrypt(message, password):
    salt = os.urandom(16)
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        iterations=100000,
        backend=default_backend()
    )
    key = base64.urlsafe_b64encode(kdf.derive(password.encode()))
    fernet = Fernet(key)
    token = fernet.encrypt(message.encode())
    return base64.urlsafe_b64encode(salt + token).decode()

def decrypt(token_text, password):
    data = base64.urlsafe_b64decode(token_text.encode())
    salt, token = data[:16], data[16:]
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        iterations=100000,
        backend=default_backend()
    )
    key = base64.urlsafe_b64encode(kdf.derive(password.encode()))
    fernet = Fernet(key)
    return fernet.decrypt(token).decode()

def main():
    # Argument parser to handle command line arguments

    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--file', help='File to encrypt', default='secret.txt')
    args = vars(parser.parse_args())




    with open(args['file'], 'r') as file:
        secret_message = file.read()
        passwd = input("Enter password for encryption: ")
        # Usage
        cipher_text = encrypt(secret_message, passwd)
        plain_text = decrypt(cipher_text, passwd)
        print(f"Cipher Text: {cipher_text}")
        print(f"Decrypted Text: {plain_text}")


if __name__ == '__main__':
    main()


#!/usr/bin/env python
import csv
import os

import gspread.utils
import yaml
from PyQt5.QtCore import Qt, pyqtSignal, QTimer, QSize
from PyQt5.QtGui import QDrag, QPixmap, QPainter
from PyQt5.QtWidgets import (QListWidget,
                             QAbstractItemView, QListWidgetItem, QMenu, QMessageBox,
                             QInputDialog, QColorDialog, QWidget, QCheckBox, QVBoxLayout, QDialog, QPushButton,
                             QDialogButtonBox)
from qrgrader.common import get_date

from qrgrader.dialogs import ButtonEditDialog, RubricEditDialog
from qrgrader.buttons import StepButton, Shortcut, Button, TextButton, StateButton, Separator, CutterButton, MultiplierButton


class FilterListWidget(QListWidget):
    score_changed = pyqtSignal(object, int)
    goto_next = pyqtSignal()
    button_or_value_changed = pyqtSignal()

    def __init__(self, rubric, **kwargs):
        super().__init__()

        self.exam_id = None
        self.config = {}
        self.scores = {}
        self.schema_dictionary = {}
        self.schema_filename = rubric.schema_filename
        name = self.schema_filename.split(".")[0]
        self.scores_filename = name + ".yaml"

        self.current_exam_id = None
        self.modified = False
        self.buttons_height = kwargs.get("buttons_height")
        self.buttons_font = kwargs.get("buttons_font")

        self.setMinimumWidth(115)

        self.populate()
        #self.load_scores()

    def load_scores(self):
        if os.path.exists(self.scores_filename):
            with open(self.scores_filename) as file:
                self.scores = yaml.safe_load(file)


    def get_selection(self):
        selected_filters = {}
        for index in range(self.count()):
            item = self.item(index)
            button: Button = self.itemWidget(item)
            cb: QCheckBox = button.layout().itemAt(0).widget()
            if cb.isChecked():
                selected_filters[button.name] = button.get_state()
        return selected_filters

    def populate(self):
        if os.path.exists(self.schema_filename):
            # Load schema
            with open(self.schema_filename, newline='') as csvfile:
                content = yaml.full_load(csvfile)
                print("loading schema:", self.schema_filename)
                self.config = content.get("config", {})
                self.schema_dictionary = content.get("buttons", {})

                # Create lateral panel and buttons
                for button_name in self.schema_dictionary:
                    button_config = self.schema_dictionary[button_name]
                    button_config["height"] = self.buttons_height
                    button_config["font"] = self.buttons_font

                    if button_config.get("type") == 'button':
                        button = StepButton(button_name, **button_config)
                        #button.score_changed.connect(self.button_clicked)  # type: ignore
                    #elif button_config.get("type") == 'text':
                    #    button = TextButton(button_name, **button_config)
                    elif button_config.get("type") == 'cutter':
                        button = CutterButton(button_name, **button_config)
                        #button.score_changed.connect(self.button_clicked)
                    elif button_config.get("type") == 'multiplier':
                        button = MultiplierButton(button_name, **button_config)
                        #button.score_changed.connect(self.button_clicked)
                    elif button_config.get("type") == 'separator':
                        button = Separator(button_name, **button_config)
                    else:
                        button = None

                    if button is not None:
                        lay: QVBoxLayout = button.layout()
                        cb = QCheckBox()
                        cb.setMaximumWidth(20)
                        lay.insertWidget(0, cb)

                        # Create Item in ListWidget
                        item = QListWidgetItem()
                        self.addItem(item)
                        self.setItemWidget(item, button)
                        item.setSizeHint(button.sizeHint())



class FilterDialog(QDialog):
    filter_changed = pyqtSignal(list)

    def __init__(self, rubric, **kwargs):
        super().__init__()

        self.setWindowTitle("Filter Questions")
        self.layout = QVBoxLayout()
        self.setLayout(self.layout)
        self.rubric = rubric

        self.filter_list = FilterListWidget(rubric, **kwargs)
        self.layout.addWidget(self.filter_list)

        # add standard OK cancel dialog buttons at the bottom

        buttons = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        self.layout.addWidget(buttons)
        QTimer.singleShot(10, self.adjustSize)

    def get_selection(self):
        return self.filter_list.get_selection()


from PyQt5 import QtWidgets, QtGui, Qt, QtCore
from PyQt5.QtCore import QTimer
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import (QLineEdit, QPushButton, QSpinBox, QDialog, QDialogButtonBox, QComboBox, QDoubleSpinBox,
                             QFormLayout, QCheckBox, QApplication, QDialog, QPushButton,
                             QGridLayout, QVBoxLayout, QHBoxLayout,
                             QDialogButtonBox, QLabel)

from qrgrader.widget_utils import WidgetsRow, VBox


class ButtonEditDialog(QDialog):
    def __init__(self, draggable_list, button=None):
        super().__init__()

        schema = button.get_config() if button is not None else {}

        self.draggable_list = draggable_list
        self.setWindowTitle("Edit")

        dialog_ok_cancel_btn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(dialog_ok_cancel_btn)
        self.buttonBox.accepted.connect(self.accept)  # type: ignore
        self.buttonBox.rejected.connect(self.reject)  # type: ignore
        self.buttonBox.buttons()[0].setEnabled(False)

        self.layout = VBox()
        self.setLayout(self.layout)

        self.combo = QComboBox()
        self.combo.addItems(['button', 'multiplier', 'cutter', 'text', 'separator'])
        self.combo.setCurrentText(schema.get('type', 'button'))
        self.layout.addWidget(WidgetsRow("Type", self.combo))

        self.le = QLineEdit()
        self.le.textChanged.connect(lambda: self.buttonBox.buttons()[0].setEnabled(self.le.text() != ""))  # type: ignore
        self.le.setText(button.get_name() if button is not None else "")
        self.layout.addWidget(WidgetsRow("Name", self.le))

        # self.value = QComboBox()
        self.value = QDoubleSpinBox()
        self.value.setDecimals(1)
        self.value.setMinimum(-20)
        self.value.setMaximum(20)
        self.value.setValue(float(schema.get('full_value', 1)))
        self.value.setSingleStep(0.5)
        self.value.valueChanged.connect(self.spin_value_changed)  # type: ignore

        # self.fill_cb(schema.get("type", 'button'), schema.get('value', 1))
        self.layout.addWidget(WidgetsRow("Value", self.value))

        self.steps = QSpinBox()
        self.steps.setMinimum(0)
        self.steps.setMaximum(10)
        self.steps.setValue(int(schema.get('steps', 0)))
        self.layout.addWidget(WidgetsRow("Steps", self.steps))
        #
        self.percent = QSpinBox()
        self.percent.setMinimum(0)
        self.percent.setMaximum(100)
        self.percent.setValue(int(schema.get('percent', 1) * 100))
        self.layout.addWidget(WidgetsRow("Percent", self.percent))
        #
        self.weight = QDoubleSpinBox()
        self.weight.setDecimals(2)
        self.weight.setMinimum(-20)
        self.weight.setMaximum(20)
        self.weight.setValue(float(schema.get('weight', 1)))
        self.weight.setSingleStep(0.5)
        self.layout.addWidget(WidgetsRow("Weight", self.weight))

        self.click_next_cb = QCheckBox("Next on click")
        self.click_next_cb.setChecked(schema.get('click_next', False))
        self.layout.addWidget(self.click_next_cb)

        # Show the current color and a color picker button
        self.color = QColor(schema.get('color', "#D4D4D4"))
        self.colorButton = QPushButton('Choose color')
        self.colorButton.clicked.connect(self.pick_color)  # type: ignore
        self.layout.addWidget(WidgetsRow('Color', self.colorButton))
        self.colorButton.setStyleSheet(f'background-color: {schema.get("color", "#D4D4D4")}')

        self.layout.addWidget(self.buttonBox)
        self.combo.currentTextChanged.connect(self.cb_changed)  # type: ignore

        self.enable_widgets()

    def spin_value_changed(self, value):
        if value < 0:
            self.weight.setStyleSheet('background-color: red')
            self.weight.setValue(0)
            QTimer.singleShot(1000, lambda: self.weight.setStyleSheet(''))

    def cb_changed(self, text):
        #        self.fill_cb(self.combo.currentText(), 1)
        self.enable_widgets()

    def enable_widgets(self):
        b = self.combo.currentText() in ['button']
        bm = self.combo.currentText() in ['button', 'multiplier']
        bmc = self.combo.currentText() in ['button', 'multiplier', 'cutter']
        mc = self.combo.currentText() in ['multiplier', 'cutter']

        self.layout.widgets['Value'].setVisible(b)
        self.layout.widgets['Steps'].setVisible(b)
        self.layout.widgets['Weight'].setVisible(b)
        self.layout.widgets['Color'].setVisible(bmc)
        self.layout.widgets['Percent'].setVisible(mc)
        self.click_next_cb.setVisible(b)

        self.adjustSize()

    def pick_color(self, button):
        color = QtWidgets.QColorDialog.getColor()
        if color.isValid():
            self.color = color
            self.colorButton.setStyleSheet(f'background-color: {color.name()}')

    def get_stylesheet(self):
        return f'background-color: {self.color.name()}'

    def get(self):
        res = {'type': self.combo.currentText()}
        if self.combo.currentText() in ['button', 'multiplier', 'cutter']:
            res['color'] = self.color.name()
        if self.combo.currentText() in ['multiplier', 'cutter']:
            res['percent'] = float(self.percent.value() / 100.0)
        if self.combo.currentText() in ['button']:
            res['steps'] = self.steps.value()
            res['weight'] = float(self.weight.value())
            res['full_value'] = float(self.value.value())
            res['click_next'] = self.click_next_cb.isChecked()

        return self.le.text(), self.combo.currentText(), res


class RubricEditDialog(QDialog):
    def __init__(self, config):
        super().__init__()
        self.config = config
        self.setWindowTitle("Edit")

        dialog_ok_cancel_btn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(dialog_ok_cancel_btn)
        self.buttonBox.accepted.connect(self.accept)  # type: ignore
        self.buttonBox.rejected.connect(self.reject)  # type: ignore
        self.buttonBox.buttons()[0].setEnabled(False)

        self.layout = QFormLayout()
        self.setLayout(self.layout)

        self.combo = QLineEdit()
        self.combo.setValidator(QtGui.QIntValidator(1, 99))
        self.layout.addRow("Page", self.combo)
        self.combo.setText(str(self.config.get("page", 1)))

        self.le = QLineEdit()

        # limit to 1 decimal
        self.le.setValidator(QtGui.QDoubleValidator(0, 10, 2))

        self.le.textChanged.connect(lambda: self.buttonBox.buttons()[0].setEnabled(self.le.text() != ""))  # type: ignore
        self.le.setText(str(self.config.get("weight", 10)))
        self.layout.addRow("Weight", self.le)

        self.precision = QLineEdit()
        self.precision.setValidator(QtGui.QIntValidator(1, 4))
        self.layout.addRow("Precision", self.precision)
        self.precision.setText(str(self.config.get("precision", 2)))


        self.layout.addWidget(self.buttonBox)

    # if accepted, modify the config
    def accept(self):
        self.config["weight"] = float(self.le.text())
        self.config["page"] = int(self.combo.text())
        self.config["precision"] = int(self.precision.text())
        super().accept()


class ControlDialog(QDialog):
    def __init__(self, page, accepting, up, down, left, right, scale_up, scale_down, parent=None):
        super().__init__(parent)
        self.setWindowTitle(f" ")
        self.setFixedSize(150, 170)
        self.accepting = accepting

        # Direction buttons
        btn_up = QPushButton("↑")
        btn_up.clicked.connect(up)
        btn_down = QPushButton("↓")
        btn_down.clicked.connect(down)
        btn_left = QPushButton("←")
        btn_left.clicked.connect(left)
        btn_right = QPushButton("→")
        btn_right.clicked.connect(right)

        # Scale buttons
        btn_scale_up = QPushButton("+")
        btn_scale_up.clicked.connect(scale_up)
        btn_scale_down = QPushButton("-")
        btn_scale_down.clicked.connect(scale_down)

        # Compact size
        for b in [btn_up, btn_down, btn_left, btn_right, btn_scale_up, btn_scale_down]:
            b.setFixedSize(28, 28)
            b.setAutoRepeat(True)
            b.setAutoRepeatInterval(25)

        grid = QGridLayout()
        #grid.setSpacing(2)
        #grid.setContentsMargins(2, 2, 2, 2)

        grid.addWidget(btn_up, 0, 1)
        grid.addWidget(btn_left, 1, 0)
        grid.addWidget(btn_right, 1, 2)
        grid.addWidget(btn_down, 2, 1)

        grid.addWidget(btn_scale_down, 0, 2)
        grid.addWidget(btn_scale_up, 2, 2)

        close = QPushButton("Done")
        close.clicked.connect(self.close)
        layout = QVBoxLayout()
        label = QLabel(f"Page {page + 1}")
        label.setAlignment(QtCore.Qt.AlignCenter)
        layout.addWidget(label)
        layout.addLayout(grid)
        # buttons.setContentsMargins(2,10,2,2)
        layout.addWidget(close)

        self.setLayout(layout)
    # on ok clicked
    def closeEvent(self, a):
        super().closeEvent(a)
        self.accepting()




import argparse
import os
import shutil
import sys
import time
from itertools import accumulate
from multiprocessing import Manager, Pool, Process
from random import randint

import cv2
import pandas as pd
import pymupdf
from PyQt5.QtCore import QObject, pyqtSignal
from PyQt5.QtWidgets import QGraphicsRectItem
from pymupdf.mupdf import PDF_ENCRYPT_KEEP

from qrgrader.code import Code
from qrgrader.code_set import CodeSet, PageCodeSet
from qrgrader.common import check_workspace, get_workspace_paths, get_temp_paths, Generated, Questions, get_date, Nia, \
    StudentsData
from qrgrader.page_processor import PageProcessor
from qrgrader.utils import makedir
from datetime import timedelta


def main():
    parser = argparse.ArgumentParser(description='Patching and detection')

    parser.add_argument('-a', '--annotate', help='Annotate files', action="store_true")
    parser.add_argument('-B', '--begin', type=int, help='First page to process', default=0)
    parser.add_argument('-c', '--correct', help='Correct QRs position (according to -x -y and -z)', action="store_true")
    parser.add_argument('-d', '--dpi', help='Dot per inch', type=int, default=400)
    parser.add_argument('-e', '--reconstruct', help='Reconstruct exams', action="store_true")
    parser.add_argument('-E', '--end', type=int, help='Last page to process', default=None)
    parser.add_argument('-g', '--export', type=str, help='Export project for storage', default=None)
    parser.add_argument('-j', '--threads', help='Number of threads to be used for processing', type=int, default=4)
    parser.add_argument('-n', '--nia', help='Create NIA file', action="store_true")
    parser.add_argument('-p', '--process', help='Options -sne', action="store_true")
    parser.add_argument('-q', '--postprocess', help='Options -nrta', action="store_true")
    parser.add_argument('-R', '--ratio', type=int, help='Resize image to save space', default=0.25)
    parser.add_argument('-r', '--raw', help='Create RAW file', action="store_true")
    parser.add_argument('-s', '--scan', help='Process pages in scanned folder', action="store_true")
    parser.add_argument('-S', '--simulate', help='Create random marked files', type=int, default=0)
    parser.add_argument('-t', '--table', help='Generate table',action="store_true")
    parser.add_argument('-T', '--temp', help='Specify temp directory', type=str, default="/tmp")
    parser.add_argument('-x', '--xdisp', help='Specify printer X displacement', type=float, default=0.0)
    parser.add_argument('-y', '--ydisp', help='Specify printer Y displacement', type=float, default=0.0)
    parser.add_argument('-z', '--zoom', help='Specify printer zoom', type=float, default=1.0)


    args = vars(parser.parse_args())

    if not check_workspace():
        print("ERROR: qrscanner must be run from a workspace directory")
        sys.exit(1)

    dir_workspace, dir_data, dir_scanned, dir_generated, dir_xls, dir_publish, dir_source = get_workspace_paths(os.getcwd())
    dir_temp_scanner, _ = get_temp_paths(get_date(), os.getcwd() if args.get("temp") is None else args.get("temp"))

    prefix = str(get_date()) + "_"
    ppm = args.get("dpi") / 25.4
    ws_date = get_date()

    if args.get("export") is not None:
        path = args.get("export")
        if not os.path.exists(path):
            print("ERROR: Export path {} does not exist".format(path))
            sys.exit(1)

        base_dir = path + os.sep + "qrgrading-" + get_date()

        print("Exporting to {}".format(base_dir))

        makedir(base_dir, clear=True)
        makedir(base_dir + os.sep + "scanned", clear=True)
        makedir(base_dir + os.sep + "generated", clear=True)
        #shutil.copytree("generated", base_dir + os.sep + "generated")
        shutil.copytree("data", base_dir + os.sep + "data")
        shutil.copytree("results", base_dir + os.sep + "results")
        shutil.copytree("source", base_dir + os.sep + "source")

        for item in os.listdir("."):
            s = os.path.join(".", item)
            if os.path.isfile(s):
                shutil.copy2(s, base_dir)
        sys.exit(0)

    if args.get("process"):
        args["scan"] = True
        args["nia"] = True
        args["reconstruct"] = True

    if args.get("postprocess"):
        args["nia"] = True
        args["raw"] = True
        args["table"] = True
        args["annotate"] = True

    zoom = args.get("zoom", 1.0)
    xdisp = args.get("xdisp", 0.0)
    ydisp = args.get("ydisp", 0.0)

    if args.get("correct") > 0:
        codes = CodeSet()
        if not codes.load(dir_data + prefix + "detected.csv"):
            print(f"ERROR: file {os.path.basename(dir_data + prefix + 'detected.csv')} not found")
            sys.exit(1)

        for code in codes:
            code: Code
            code.x = code.x * zoom + xdisp
            code.y = code.y * zoom + ydisp
            code.w = code.w * zoom
            code.h = code.h * zoom

        codes.save(dir_data + prefix + "detected.csv")


    if (simulate := args.get("simulate")) > 0:
        print("Simulation in progress ({} files)".format(simulate))
        makedir(dir_scanned, clear=True)

        generated = Generated(72 / 25.4)

        if not generated.load(dir_data + prefix + "generated.csv"):
            print(f"ERROR: file {os.path.basename(dir_data + prefix + 'generated.csv')} not found")
            sys.exit(1)

        pdf_filenames = [f for f in os.listdir(dir_generated) if f.endswith(".pdf")]
        pdf_filenames.sort()

        for pdf_filename in pdf_filenames[0:simulate]:
            print("Marking random answers in {}".format(pdf_filename), end="\r")

            new_pdf = pymupdf.open()
            doc = pymupdf.open(dir_generated + pdf_filename)

            exam = pdf_filename[6:9]

            for page in doc:

                filtered = generated.select(type=Code.TYPE_A, exam=int(exam), page=page.number + 1)
                for question in filtered.get_questions():
                    qrs = filtered.select(type=Code.TYPE_A, question=question)
                    for qr in qrs:
                        if qr.answer == randint(1, 4):
                            x = qr.x + 5
                            y = qr.y - 7
                            w = 10
                            h = 10
                            annot = page.add_redact_annot(pymupdf.Rect(x, y, x + w, y + h), fill=(0.5, 0.5, 0.5), cross_out=False)
                            break

                nias = generated.select(type=Code.TYPE_N, exam=int(exam), page=page.number + 1)
                for i in range(6):
                    r = randint(0, 9)
                    qr = nias.first(number=i * 10 + r)
                    if qr is not None:
                        x = qr.x + 5
                        y = qr.y - 7
                        w = 10
                        h = 10
                        annot = page.add_redact_annot(pymupdf.Rect(x, y, x + w, y + h), fill=(0.5, 0.5, 0.5), cross_out=False)

                # Apply the redactions
                page.apply_redactions()
                # Get the images with the marked codes
                pix = page.get_pixmap(matrix=pymupdf.Matrix(args.get("dpi") / 72, args.get("dpi") / 72))
                # Insert the new page and the image inside it
                new_page = new_pdf.new_page()
                new_page.insert_image(new_page.rect, stream=pix.tobytes("jpg"), width=pix.width, height=pix.height)
                # To be sure avoiding memory leaks
                del pix

            new_pdf.save(dir_scanned + pdf_filename)
            new_pdf.close()
            doc.close()

        print("\nSimulation done.")

    if args.get("scan", False):
        os.makedirs(dir_temp_scanner, exist_ok=True)

        first_page = args.get("begin")
        last_page = args.get("end")

        generated = Generated(ppm)
        if not generated.load(dir_data + prefix + "generated.csv"):
            print(f"ERROR: file {os.path.basename(dir_data + prefix + 'generated.csv')} not found")
            sys.exit(1)

        files = []
        for i, filename in enumerate(sorted([x for x in os.listdir(dir_scanned) if x.endswith(".pdf")])):
            document = pymupdf.open(dir_scanned + filename)
            files.append((i, filename, first_page, len(document) if last_page is None else last_page))
            document.close()

        total_length = sum(length-first for _, _, first, length in files)
        if total_length <= 0:
            print("No pages to process. Exiting.")
            sys.exit(0)
        time_begin = time.time()
        time_remaining = 0
        elapsed = 0

        with Manager() as manager:

            done = 0
            processes = []
            detected = manager.list()
            semaphore = manager.BoundedSemaphore(args.get("threads"))

            for index, filename, first, length in files:

                index > 0 and print() # for the \r at the end of the last line
                print(f">> Processing file {filename} ({index+1}/{len(files)})")

                for i in range(first, length):

                    semaphore.acquire()
                    done += 1

                    procs = [p for p in processes if not p.is_alive()].copy()
                    for process in procs:
                        process.join()
                        processes.remove(process)

                    # We send the filename and open the document in the process for three reasons:
                    # 1. Sending the page object is not possible because it is not pickable
                    # 2. Rendering the page image in parallel make the whole process much faster
                    # 3. For some reason sending the image to the process creates memory overflow

                    process = PageProcessor(semaphore, dir_scanned + filename, i, generated, detected, dir_images=dir_temp_scanner, resize=args.get("ratio"))
                    processes.append(process)
                    process.start()

                    #while len([p for p in processes if p.is_alive()]) >= 4:
                    #    time.sleep(0.25)

                    time_remaining = (time.time() - time_begin) / (done+1) * (total_length - done)
                    h, r = divmod(int(time_remaining), 3600)
                    m, s = divmod(r, 60)

                    print(f"   Processed {done}/{total_length} ({100*done/total_length:.1f}%) ({len(detected)} codes found) remaining: {h:02}:{m:02}:{s:02}", end="\r")

            print() # for the \r at the end of the last line

            for process in processes:
                process.join()

            codes = CodeSet()
            codes.extend(detected)
            codes.save(dir_data + prefix + "detected.csv")

    if args.get("reconstruct") or args.get("nia") or args.get("raw") or args.get("annotate"):
        codes = CodeSet()
        if not codes.load(dir_data + prefix + "detected.csv"):
            print(f"ERROR: file {os.path.basename(dir_data + prefix + 'detected.csv')} not found")
            sys.exit(1)

        exams = codes.get_exams()
        date = codes.get_date()

    if args.get("reconstruct"):
        print(">> Reconstructing exams")
        images = os.listdir(dir_temp_scanner)
        for exam in exams:
            filename = dir_publish + "{}{:03d}.pdf".format(date, exam)
            pdf_file = pymupdf.open()
            exam_images = sorted([x for x in images if x.startswith("page-{}-{}-".format(date, exam))])
            for image in exam_images:
                page = pdf_file.new_page()  # noqa
                page.insert_image(pymupdf.Rect(0, 0, 595.28, 842), filename=dir_temp_scanner + os.sep + image)

            pdf_file.save(filename)

    if args.get("nia"):
        nia_filename = dir_xls + prefix + "nia.csv"
        print(f">> Creating {os.path.basename(nia_filename)} file")
        type_n = codes.select(type=Code.TYPE_N)
        with open(nia_filename, "w", encoding='utf-8') as f:
            f.write("EXAM\tNIA\n")
            for exam in exams:
                nia = {0: 'Y', 1: 'Y', 2: 'Y', 3: 'Y', 4: 'Y', 5: 'Y'}
                exam_codes = type_n.select(exam=exam)
                for row in range(6):
                    for number in range(10):
                        result = exam_codes.first(number=row * 10 + number)
                        if result is None or result.marked:
                            nia[row] = number if nia[row] == 'Y' else 'X'
                nia = "".join([str(x) for x in nia.values()])

                f.write("{}\t{}\n".format(date * 1000 + exam, nia))

    if args.get("raw"):
        raw_filename = dir_xls + prefix + "raw.csv"
        print(f">> Creating {os.path.basename(raw_filename)} file")
        with open(raw_filename, "w", encoding='utf-8') as f:
            # # Header
            # line = "DATE\tEXAM"
            # for qn in codes.get_questions():
            #     for an in codes.get_answers():
            #         line += "\tQ{:d}{}".format(qn, chr(64 + an))
            # for on in codes.get_open():
            #     line += "\tO{:d}".format(on)
            # f.write(line + "\n")

            # Exams
            type_a = codes.select(type=Code.TYPE_A)
            type_o = codes.select(type=Code.TYPE_O)
            questions = type_a.get_questions()
            answers = type_a.get_answers()
            openq = type_o.get_open()

            # print("questions: {}".format(questions))
            # print("answers: {}".format(answers))
            # print("openq: {}".format(openq))

            for exam in exams:
                exam_codes = type_a.select(exam=exam)
                line = f"{date}\t{exam}"
                for question in questions:
                    for answer in answers:
                        result = exam_codes.first(question=question, answer=answer)
                        line += "\t1" if result is None or result.marked else "\t0"

                for _ in openq:
                    line += "\t0"
                f.write(line + "\n")
    if args.get("table"):
        table_filename = dir_xls + prefix + "table.csv"
        print(f">> Creating {os.path.basename(table_filename)} file")

        nia = Nia(dir_xls + prefix + "nia.csv")
        nia.load()

        students_data = StudentsData(dir_xls + os.sep + "data.csv")
        students_data.load()

        questions = Questions(dir_xls + os.sep + str(ws_date) + "_questions.csv")
        questions.load()

        raw = pd.read_csv(dir_xls + os.sep + str(ws_date) + "_raw.csv", sep='\t', header=None)
        raw.iloc[:, 0] = raw.iloc[:, 0] * 1000 + raw.iloc[:, 1]
        raw.set_index(raw.iloc[:, 0], inplace=True)

        hdr = 4  # Number of header rows
        inserted = 0

        # Insert the NIA column
        raw.insert(inserted + 2, "NIA", "")
        for exam_id in raw.iloc[:, 0].tolist():
            raw.loc[exam_id, "NIA"] = nia.get_nia(exam_id)
        inserted += 1

        # Insert the GRADE columns
        raw.insert(inserted + 2, "T", '=INDIRECT(ADDRESS(ROW(), COLUMN() -2)) + INDIRECT(ADDRESS(ROW(), COLUMN() -1))')
        raw.insert(inserted + 2, "O", '=SUMPRODUCT(($G$3:$3="O") * INDIRECT(ADDRESS(ROW(), COLUMN() + 2) & ":" '
                                      '& ADDRESS(ROW(), COLUMNS($A$1:$1)))* INDIRECT(ADDRESS(4, COLUMN() + 2) & ":" '
                                      '& ADDRESS(4, COLUMNS($A$1:$1))))')
        raw.insert(inserted + 2, "Q", '=max(0,SUMPRODUCT(($G$3:$3<>"O") *INDIRECT(ADDRESS(ROW(), COLUMN() + 3) & ":" '
                                      '& ADDRESS(ROW(), COLUMNS($A$1:$1)))* INDIRECT(ADDRESS(4, COLUMN() + 3) & ":" '
                                      '& ADDRESS(4, COLUMNS($A$1:$1)))))')

        inserted += 3

        # Fill the header
        percent_answ = '=SUM(OFFSET(INDIRECT("RC", FALSE), 1, 0, ROWS(A:A)-ROW(), 1))/COUNT(OFFSET(INDIRECT("RC", FALSE), 1, 0, ROWS(A:A)-ROW(), 1))'
        percent_ques = ('=SUM(OFFSET(INDIRECT("RC", FALSE), ' + str(
            hdr) + ', -3, ROWS(A:A) - ROW()-4, 4))/COUNT(OFFSET(INDIRECT("RC", FALSE), '
                        + str(hdr) + ', 0, ROWS(A:A)-ROW(), 1))')

        names, qn, ans_letter, ans_value = [""] * (inserted + 2), [""] * (inserted + 2), [""] * (inserted + 2), [
            ""] * (inserted + 2)
        ans_perc = ["Exam ID", "#", "NIA", "Q", "O", "T"]

        num_open = 0
        for question in questions.get_questions():
            if questions.get_type(question) == "Q":
                names.extend([questions.get_text(question), "", "", percent_ques])
                qn.extend([question, question, question, question])
                ans_value.extend([questions.get_value(question, i + 1) for i in range(4)])
                ans_letter.extend([chr(65 + i) for i in range(4)])
                ans_perc.extend([percent_answ] * 4)
            else:
                names.append(questions.get_text(question))
                qn.append(question)
                ans_letter.append("O")
                ans_value.append(questions.get_value(question, 1))
                ans_perc.append('=countif((INDIRECT(ADDRESS(6, COLUMN()) & ":" & ADDRESS(ROWS(6:1000), COLUMN()))),">0")/count((INDIRECT(ADDRESS(6, COLUMN()) & ":" & ADDRESS(ROWS(6:1000), COLUMN()))))')
                num_open += 1

        # print(raw.shape, len(names), len(qn), len(ans_letter), len(ans_value), len(ans_perc))

        raw.loc[-5] = names
        raw.loc[-4] = qn
        raw.loc[-3] = ans_letter
        raw.loc[-2] = ans_value
        raw.loc[-1] = ans_perc
        df = raw.sort_index().reset_index(drop=True)

        df.iloc[5:,-num_open:] = '=IFERROR(VLOOKUP(INDIRECT("A" & ROW()), INDIRECT("'+get_date()+'_" & INDIRECT(ADDRESS(1, COLUMN(), 4)) & "!A:D"), 2, FALSE),0)'

        df.to_csv(table_filename, sep='\t', index=False, header=False)


    if args.get("annotate"):
        print(">> Annotating exams")
        questions = Questions(dir_xls + prefix + "questions.csv")
        if not questions.load():
            print(f"ERROR: file {os.path.basename(dir_xls + prefix + 'questions.csv')} not found")
            sys.exit(1)

        for exam in exams:
            print("   Annotating exam {}".format(exam), end="\r")
            filename = dir_publish + "{}{:03d}.pdf".format(date, exam)
            pdf_file = pymupdf.open(filename)
            for page in pdf_file:

                for annot in page.annots():
                    page.delete_annot(annot)

                this_page = codes.select(type=Code.TYPE_A, exam=exam, page=page.number + 1)
                for code in this_page:
                    if code.marked:
                        x, y = code.get_pos()
                        w, h = code.get_size()
                        r = pymupdf.Rect(xdisp + x*zoom, ydisp + y*zoom, xdisp + x*zoom + w*zoom, ydisp + y*zoom + h*zoom)
                        annot = page.add_rect_annot(r)
                        annot.set_border(width=2)
                        if questions.get_value(code.question, code.answer) > 0:
                            annot.set_colors(stroke=(0, 1, 0))
                        else:
                            annot.set_colors(stroke=(1, 0, 0))
                        annot.update()
            pdf_file.save(filename, incremental=True, encryption=PDF_ENCRYPT_KEEP)
        print()

    print("All done :)")


if __name__ == '__main__':
    main()



import argparse
import math
import os
import sys
import time
from multiprocessing import Manager, Pool, Process

import cv2
import numpy as np
import pymupdf

from qrgrader.code import Code
from qrgrader.code_set import CodeSet, PageCodeSet
from qrgrader.utils import pix2np, get_patches, threshold, get_codes, compute_similarity_transform, \
    get_similarity_transform


class PageProcessor(Process):

    def __init__(self, semaphore, filename, index, generated, result, **kwargs):
        super().__init__()
        self.filename = filename
        self.index = index
        self.generated = generated
        self.result = result
        self.semaphore = semaphore

        self.dpi = kwargs.get("dpi", 400)
        self.thresholds = kwargs.get("thresholds", [50, 55, 60, 65, 70, 75, 80])
        self.matrix = pymupdf.Matrix(self.dpi / 72, self.dpi / 72)
        self.show_patches = kwargs.get("show_patches", False)
        self.resize = kwargs.get("resize", 1.0)
        self.dir_images = kwargs.get("dir_images", "../..")

        self.ppm = self.dpi / 25.4

    def run(self):

        # Render the image
        doc = pymupdf.open(self.filename)
        page = doc[self.index]
        image = pix2np(page.get_pixmap(matrix=self.matrix))  # noqa
        doc.close()


        # Find page, orientation and rotate page
        rotation = None
        detected = PageCodeSet()
        h, w = image.shape[:2]

        for th in self.thresholds:

            ne = threshold(image[0:500, w - 500:w], th)
            for text, cx, cy, cw, ch in get_codes(ne):
                if text.startswith("P"):
                    rotation = -1
                elif text.startswith("Q"):
                    rotation = cv2.ROTATE_180
                detected.append(Code(text, cx, cy, cw, ch))

            if rotation is None:
                nw = threshold(image[0:500, 0:500], th)
                for text, cx, cy, cw, ch in get_codes(nw):
                    if text.startswith("P"):
                        rotation = cv2.ROTATE_90_CLOCKWISE
                    elif text.startswith("Q"):
                        rotation = cv2.ROTATE_90_COUNTERCLOCKWISE
                    detected.append(Code(text, cx, cy, cw, ch))

            if rotation is None:
                sw = threshold(image[h - 500:h, 0:500], th)
                for text, cx, cy, cw, ch in get_codes(sw):
                    if text.startswith("P"):
                        rotation = cv2.ROTATE_180
                    elif text.startswith("Q"):
                        rotation = -1
                    detected.append(Code(text, cx, cy, cw, ch))

            if rotation is None:
                se = threshold(image[h - 500:h, w - 500:w], th)
                for text, cx, cy, cw, ch in get_codes(se):
                    if text.startswith("P"):
                        rotation = cv2.ROTATE_90_COUNTERCLOCKWISE
                    elif text.startswith("Q"):
                        rotation = cv2.ROTATE_90_CLOCKWISE
                    detected.append(Code(text, cx, cy, cw, ch))

            if rotation is not None:
                break

        if rotation is not None and rotation != -1:
            image = cv2.rotate(image, rotation)

        # Get the page number if we got it
        page = detected.get_page()

        # Clear the detected because the
        # page may have been rotated
        detected.clear()

        # Process the page and extract the detected
        for th in self.thresholds:
            th_image = threshold(image, th)
            patches = get_patches(th_image, self.ppm, 8)

            for px, py, pw, ph in patches:
                patch = image[py:py + ph, px:px + pw]

                if self.show_patches:
                    cv2.rectangle(image, (px, py), (px + pw, py + ph), (0, 0, 255), 1)

                for text, cx, cy, cw, ch in get_codes(patch):
                    detected.append(Code(text, px + cx, py + cy, cw, ch, page, self.index))

        # Try again with the whole page
        page = detected.get_page() if page is None else page
        exam = detected.get_exam_id()

        # If we did not find the page number, try to find it in the generated detected
        if page is None:
            if len(detected) > 0:
                page = self.generated.get(detected.first()).page
                for code in detected:
                    code.set_page(page)

        if self.resize != 1.0:
            image = cv2.resize(image, (int(image.shape[1] * self.resize), int(image.shape[0] * self.resize)),
                               interpolation=cv2.INTER_AREA)

        if page is not None:
            cv2.imwrite(self.dir_images + os.sep + "page-{}-{}-{:03d}.jpg".format(detected.get_date(), detected.get_exam_id(), page), image)
        elif detected.get_exam_id():
            cv2.imwrite(self.dir_images + os.sep + "page-{}-{}-{:03d}.jpg".format(detected.get_date(), detected.get_exam_id(), 0), image)
        else:
            cv2.imwrite(self.dir_images + os.sep + "{}-{:03d}.jpg".format(self.filename, self.index), image)

        # # Compute the transformation
        generated_page_codeset = PageCodeSet(self.generated.select(exam=exam, page=page))

        # We will use the two codes that are the farthest apart to compute
        # the transformation, if we have at least two codes detected
        max_dist, codes = 0, None
        for code1 in detected:
            for code2 in detected:
                if code1!=code2:
                    dist = math.sqrt((code1.get_pos()[0] - code2.get_pos()[0]) ** 2 + (code1.get_pos()[1] - code2.get_pos()[1]) ** 2)
                    if dist > max_dist:
                        max_dist = dist
                        codes = (code1, code2)

        if codes is not None:
            # we have at least two codes detected, we can use them to compute the transformation
            code1, code2 = codes[0], codes[1]
            p11 = self.generated.get(code1).get_pos()
            p12 = self.generated.get(code2).get_pos()
            p21 = code1.get_pos()
            p22 = code2.get_pos()
            transform = get_similarity_transform([p11, p12], [p21, p22]) # noqa
        elif len(detected) > 0:
            # we have detected just one code, we can use it to compute the transformation
            code1 = detected.first()
            p11 = self.generated.get(code1).get_pos()
            p21 = code1.get_pos()
            transform = lambda pt: (pt[0] + (p21[0] - p11[0]), pt[1] + (p21[1] - p11[1]))
        else:
            # No codes detected, we can not compute the transformation, we will just use the identity
            transform = lambda pt: pt

        # TODO: NOTHING TO DO, JUST A NOTE
        # TODO: ATTENTION! This is the KEY: ALL THE CODES will be present in the detected set,
        # TODO: but they will be marked as "marked" if they were not detected in the page.
        # TODO: This way we can keep track of all the codes and their positions, even if
        # TODO: they were not detected in the page.

        for code in generated_page_codeset:
            new_pose = transform(code.get_pos())
            code.set_pos(new_pose)  # Note: OpenCV uses (y, x) order
            code.set_size(120, 120)
            code.scale(72.0 / self.dpi)
            code.set_marked(detected.get(code) is None)
            self.result.append(code)

        #print(f"Processed {os.path.basename(self.filename)} page {self.index} ({len(generated_page_codeset)} codes detected)")
        self.semaphore.release()



import argparse
import os
import re
import sys

import yaml

from qrgrader.common import check_workspace, Nia, get_workspace_paths, get_date
from qrgrader.encrypt import decrypt
from qrgrader.secret import get_secret
from qrgrader.utils import makedir

# sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from qrgrader.gdrive import GDrive, Sheets

client_secrets_json = '''
{

}
'''


def main():
    parser = argparse.ArgumentParser(description='Upload and download sheets from Google Sheets')
    parser.add_argument('-d', '--download', help='Download sheet', action="append", default=[])
    parser.add_argument('-D', '--download-all', help='Download sheet', action="store_true")
    parser.add_argument('-f', '--file', help='Config file')
    parser.add_argument('-F', '--filter', help='Filter names')
    parser.add_argument('-g', '--upload-pdf', help='Specify folder id for uploading', action="store_true")
    parser.add_argument('-i', '--folder-id', help='Specify folder id for uploading', default=None)
    parser.add_argument('-l', '--list', help='Create a PDF sheet', action="store_true")
    parser.add_argument('-p', '--folder-path', help='Specify folder path for uploading', default=None)
    parser.add_argument('-P', '--shared-folder-name', help='Specify folder name for uploading or listing', default=None)
    parser.add_argument('-u', '--upload', help='Upload sheet', action="append", default=[])
    parser.add_argument('-U', '--upload-all', help='Upload all sheets', action="store_true")
    parser.add_argument('-w', '--workbook', help='Workbook name')
    parser.add_argument('-x', '--diff', help='Difference')
    parser.add_argument('-y', '--yeah', help="Answer 'yes' to all questions", action="store_true")
    args = vars(parser.parse_args())

    if not check_workspace():
        print("ERROR: qrsheets must be run from a workspace directory")
        sys.exit(1)

    dir_workspace, dir_data, _, dir_generated, dir_xls, _, dir_source = get_workspace_paths(os.getcwd())
    date = get_date()

    if not check_workspace():
        print("ERROR: qrsheets must be run from a workspace directory")
        sys.exit(1)

    # Create json file with client secrets
    makedir("config")
    if not os.path.exists("config" + os.sep + "client_secret.json"):
            secret = get_secret()
            passwd = input("Enter password for QRGrader secret: ")
            try:
                client_secrets_json = decrypt(secret, passwd)
                with open("config" + os.sep + "client_secret.json", "w", encoding='utf-8') as f:
                    f.write(client_secrets_json)
            except Exception as e:
                print("Password incorrect. You may request a password to dantard@unizar.es", e)
                sys.exit(1)

    # Load config file if provided
    if args["file"] is not None:
        with open(args["file"], "r", encoding='utf-8') as f:
            file_args = yaml.safe_load(f)
            args.update(file_args)

    if args["file"] is None and args["workbook"] is None and args["upload_pdf"] is False and args["list"] is False:
        print("You must specify at least one of the following options: --file, --workbook, --upload-pdf, --list")
        return

    if args.get("upload_pdf") or args.get("list"):

        gd = GDrive(config_dir="config")

        if args["folder_id"] is not None:
            folder_id = args["folder_id"]
        elif args["folder_path"] is not None:
            folder_id = gd.get_folder_id_by_path(args["folder_path"])
        elif args["shared_folder_name"] is not None:
            folder_id = gd.get_shared_folder_id(args["shared_folder_name"])
        else:
            print("You must specify a folder-id, folder-path or shared-folder-name")
            sys.exit(1)

        if folder_id is None:
            print("Folder does not exist, exiting.")
            sys.exit(1)

        if args.get("upload_pdf"):
            for file in os.listdir("results" + os.sep + "pdf"):
                if file.endswith(".pdf"):
                    print("Uploading file", file)
                    gd.upload_file("results" + os.sep + "pdf" + os.sep + file, folder_id)
            print("All files uploaded to folder", folder_id)
            args["list"] = folder_id

        if args.get("list", None):

            nias = Nia(dir_xls + os.sep + str(date) + "_nia.csv")
            nias.load()

            files = gd.ls(folder_id)

            files.sort(key=lambda x: x[0].lower())

            with open("results" + os.sep + "xls" + os.sep + date + "_pdf.csv", "w", encoding='utf-8') as f:
                f.write("EXAM ID" + "\t" + "NIA" + "\t" + "LINK" + "\t" + "QUIZ" + "\t" + "OPEN" + "\t" + "TOTAL" + "\n")
                for i, file in enumerate(files):
                    name = file[0].replace(".pdf", "")
                    link = "https://drive.google.com/u/3/uc?id=" + file[1]
                    nia = nias.get_nia(name)
                    fb_quiz = "=round(VLOOKUP(A" + str(i+2) + ",'"+date+"_raw'!A:Z,4,false),2)"
                    fb_open = "=round(VLOOKUP(A" + str(i + 2) + ",'"+date+"_raw'!A:Z,5,false),2)"
                    fb_total = "=round(VLOOKUP(A" + str(i + 2) + ",'"+date+"_raw'!A:Z,6,false),1)"

                    f.write(name + "\t" + str(nia) + "\t" + link + "\t" +
                            fb_quiz + "\t" + fb_open + "\t" + fb_total + "\n")

            print("Written to results/xls/{}_pdf.csv ({} rows)".format(date, len(files)))

    args_workbook = args.get("workbook", None)
    if args_workbook is not None:

        args_upload = args.get("upload", [])
        args_filter = args.get("filter", None)
        args_download = args.get("download", [])
        args_download_all = args.get("download_all", False)
        args_yes = args.get("yeah")
        args_upload_all = args.get("upload_all", False)
        args_diff = args.get("diff", None)

        # args_filter = date if args_filter is None else args_filter
        args_filter = None if args_filter == "." else args_filter

        if len(args_upload) > 0 and args_upload_all:
            print("You cannot use both --upload and --upload-all")
            return

        if len(args_download) > 0 and args_download_all:
            print("You cannot use both --download and --download-all")
            return

        sh = Sheets(base_folder="results/xls", config_dir="config")
        sh.open(args_workbook)

        if args_upload_all:
            sh.upload_all(args_filter, args_yes)

        elif len(args_upload) > 0:
            sh.upload(args_upload, args_yes)

        if args_download_all:
            sh.download_all(args_filter, args_yes)
        elif len(args_download) > 0:
            sh.download(args_download, args_yes)

        if args_diff:
            sh.diff(args_diff)

    print("All Done :)")


if __name__ == "__main__":
    main()
